"""Add progress description

Revision ID: c13116948b54
Revises: e798462e0c03
Create Date: 2025-01-07 10:35:23.654928

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "c13116948b54"
down_revision = "e798462e0c03"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("internal_jobs", sa.Column("progress_description", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("internal_jobs", "progress_description")
    # ### end Alembic commands ###
