"""add user session table

Revision ID: 521464fa93ca
Revises: 2c3c406a2750
Create Date: 2025-03-23 16:17:36.143307

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "521464fa93ca"
down_revision = "2c3c406a2750"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "user_session",
        sa.Column("session_id", sa.String(), nullable=False),
        sa.Column("session_data", sa.LargeBinary(), nullable=False),
        sa.Column("last_accessed", sa.TIMESTAMP(timezone=True), nullable=False),
        sa.PrimaryKeyConstraint("session_id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("user_session")
    # ### end Alembic commands ###
